/* IMPORT */
import makeNakedPromise from 'promise-make-naked';
import Worker from './worker/index.js';
/* MAIN */
class WorkTank {
    /* CONSTRUCTOR */
    constructor(options) {
        this.terminated = true;
        this.timeout = options.timeout ?? Infinity;
        this.terminateTimeout = options.autoterminate ?? 60000;
        this.name = options.name ?? 'WorkTank-Worker';
        this.size = options.size ?? 1;
        this.methods = this._getMethods(options.methods);
        this.tasksBusy = new Set();
        this.tasksReady = new Set();
        this.workersBusy = new Set();
        this.workersReady = new Set();
    }
    /* HELPERS */
    _autoterminate() {
        if (this.terminateTimeoutId)
            return;
        if (!this.tasksBusy.size && !this.tasksReady.size) {
            this.terminateTimeoutId = undefined;
            this.terminate();
        }
        else {
            const timeout = Math.min(2147483647, this.terminateTimeout);
            this.terminateTimeoutId = setTimeout(() => {
                this.terminateTimeoutId = undefined;
                this._autoterminate();
            }, timeout);
        }
    }
    _getMethods(methods) {
        if (typeof methods === 'string') { // Already serialized methods, or URL to import, useful for complex and/or bundled workers
            if (/^(file|https?):\/\//.test(methods)) { // URL to import
                const methodsImport = `${'import'} * as Methods ${'from'} '${methods}';`;
                const methodsRegister = `Object.entries ( Methods ).forEach ( ([ name, fn ]) => typeof fn === 'function' && WorkTankWorkerBackend.register ( name, fn ) );`;
                return `${methodsImport}\n\n${methodsRegister}`;
            }
            else { // Serialized methods
                return methods;
            }
        }
        else if (methods instanceof URL) { // URL to import
            return this._getMethods(methods.href);
        }
        else { // Serializable methods
            const names = Object.keys(methods);
            const values = Object.values(methods);
            const serialized = names.map((name, index) => `WorkTankWorkerBackend.register ( '${name}', ${values[index].toString()} );`).join('\n');
            return serialized;
        }
    }
    _getTaskReady() {
        for (const task of this.tasksReady)
            return task;
    }
    _getWorkerName() {
        if (this.size < 2)
            return this.name;
        const counter = 1 + (this.workersBusy.size + this.workersReady.size);
        return `${this.name} (${counter})`;
    }
    _getWorkerReady() {
        for (const worker of this.workersReady)
            return worker;
        if (this.workersBusy.size >= this.size)
            return;
        const name = this._getWorkerName();
        const worker = new Worker(this.methods, name);
        this.workersReady.add(worker);
        return worker;
    }
    /* API */
    exec(method, args) {
        const { promise, resolve, reject } = makeNakedPromise();
        const task = { method, args, promise, resolve, reject };
        this.terminated = false;
        this.tasksReady.add(task);
        this.tick();
        this._autoterminate();
        return promise;
    }
    proxy() {
        return new Proxy({}, {
            get: (_, method) => {
                if (typeof method !== 'string')
                    throw new Error('Unsupported method name');
                return (...args) => {
                    return this.exec(method, args);
                };
            }
        });
    }
    terminate() {
        this.terminated = true;
        /* RESETTING AUTO-TERMINATE */
        clearTimeout(this.terminateTimeoutId);
        this.terminateTimeoutId = undefined;
        /* RESETTING TASKS */
        const error = new Error('WorkTank terminated!');
        for (const task of this.tasksBusy)
            task.reject(error);
        for (const task of this.tasksReady)
            task.reject(error);
        this.tasksBusy = new Set();
        this.tasksReady = new Set();
        /* RESETTING WORKERS */
        for (const worker of this.workersBusy)
            worker.terminate();
        for (const worker of this.workersReady)
            worker.terminate();
        this.workersBusy = new Set();
        this.workersReady = new Set();
    }
    tick() {
        const worker = this._getWorkerReady();
        if (!worker)
            return;
        const task = this._getTaskReady();
        if (!task)
            return;
        this.workersReady.delete(worker);
        this.workersBusy.add(worker);
        this.tasksReady.delete(task);
        this.tasksBusy.add(task);
        let timeoutId;
        if (this.timeout > 0 && this.timeout !== Infinity) {
            const timeout = Math.min(2147483647, this.timeout);
            timeoutId = setTimeout(() => {
                worker.terminate();
            }, timeout);
        }
        task.promise.finally(() => {
            clearTimeout(timeoutId);
            if (this.terminated)
                return;
            this.workersBusy.delete(worker);
            if (!worker.terminated) {
                this.workersReady.add(worker);
            }
            this.tasksBusy.delete(task);
            this.tick();
        });
        worker.exec(task);
    }
}
/* EXPORT */
export default WorkTank;
